/**
 * ntfs.c - Internal support routines for NTFS-based devices.
 *
 * Copyright (c) 2021 Anton "Ant-ON"
 * Copyright (c) 2010 Dimok
 * Copyright (c) 2009 Rhys "Shareese" Koedijk
 * Copyright (c) 2006 Michael "Chishm" Chisholm
 *
 * This program/include file is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program/include file is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdlib.h>
#include <errno.h>
#include <string.h>

#include "ntfs.h"

#include <volume.h>
#include <inode.h>
#include <unistr.h>
#include <dir.h>
#include <reparse.h>

#define NTFS_MAX_SYMLINK_DEPTH              10 /* Maximum search depth when resolving symbolic links */


static ntfs_inode *ntfs_inode_parse(ntfs_volume *vol, const char *path, int reparseLevel)
{
    ntfs_inode *ni = NULL;
    char *target = NULL;

    // Sanity check
    if (!vol)
    {
        errno = ENODEV;
        return NULL;
    }

    // Get the actual path of the entry
    if (path[0] != '/')
        return NULL;

    // Find the entry, taking into account our current directory (if any)
    ni = ntfs_pathname_to_inode(vol, NULL, path);

    // If the entry was found and it has reparse data then parse its true path;
    // this resolves the true location of symbolic links and directory junctions
    if (ni && (ni->flags & FILE_ATTR_REPARSE_POINT))
    {
        if (ntfs_possible_symlink(ni))
        {
            // Sanity check, give up if we are parsing to deep
            if (reparseLevel > NTFS_MAX_SYMLINK_DEPTH)
            {
                ntfs_inode_close(ni);
                errno = ELOOP;
                return NULL;
            }

            // Get the target path of this entry
            target = ntfs_make_symlink(ni, path);
            if (!target)
            {
                ntfs_inode_close(ni);
                return NULL;
            }

            // Close the entry (we are no longer interested in it)
            ntfs_inode_close(ni);

            // Parse the entries target
            ni = ntfs_inode_parse(vol, target, reparseLevel++);

            // Clean up
            // use free because the value was not allocated with malloc
            free(target);

        }
    }

    return ni;
}

ntfs_inode *ntfs_inode_open_p(ntfs_volume *vol, const char *path)
{
    return ntfs_inode_parse(vol, path, 1);
}

ntfs_inode *ntfs_inode_create_p(ntfs_volume *vol, const char *path, mode_t type, const char *target)
{
    ntfs_inode *dir_ni = NULL, *ni = NULL;
    char *dir = NULL;
    char *name = NULL;
    ntfschar *uname = NULL, *utarget = NULL;
    int uname_len, utarget_len;

    // Sanity check
    if (!vol)
    {
        errno = ENODEV;
        return NULL;
    }
    if (!path)
    {
        errno = EINVAL;
        return NULL;
    }

    // Get the unicode name for the entry and find its parent directory
    // TODO: This looks horrible, clean it up
    dir = strdup(path);
    if (!dir)
    {
        errno = EINVAL;
        goto cleanup;
    }
    name = strrchr(dir, '/');
    if (name)
        name++;
    else
        name = dir;
    uname_len = ntfs_mbstoucs(name, &uname);
    if (uname_len < 0)
    {
        errno = EINVAL;
        goto cleanup;
    }
    name = strrchr(dir, '/');
    if(name)
    {
        name++;
        name[0] = 0;
    }

    // Open the entries parent directory
    dir_ni = ntfs_inode_open_p(vol, dir);
    if (!dir_ni)
        goto cleanup;

    // Create the entry
    switch (type) 
    {
        // Symbolic link
        case S_IFLNK:
            if (!target) {
                errno = EINVAL;
                goto cleanup;
            }
            utarget_len = ntfs_mbstoucs(target, &utarget);
            if (utarget_len < 0) {
                errno = EINVAL;
                goto cleanup;
            }
            ni = ntfs_create_symlink(dir_ni, 0, uname, uname_len,  utarget, utarget_len);
            break;

        // Directory or file
        case S_IFDIR:
        case S_IFREG:
            ni = ntfs_create(dir_ni, 0, uname, uname_len, type);
            break;

    }

    // If the entry was created
    if (ni)
    {
        // Mark the entry for archiving
        ni->flags |= FILE_ATTR_ARCHIVE;

        // Mark the entry as dirty
        NInoSetDirty(ni);

        // Sync the entry to disc
        ntfs_inode_sync(ni);

        // Update parent directories times
        ntfs_inode_touch(dir_ni, NTFS_UPDATE_MCTIME);

    }

cleanup:

    if(dir_ni)
        ntfs_inode_close(dir_ni);
    if(utarget)
        free(utarget);
    if(uname)
        free(uname);
    if(dir)
        free(dir);
    return ni;
}

int ntfs_link_p(ntfs_volume *vol, const char *old_path, const char *new_path)
{
    ntfs_inode *dir_ni = NULL, *ni = NULL;
    char *dir = NULL;
    char *name = NULL;
    ntfschar *uname = NULL;
    int uname_len;
    int res = 0;

    // Sanity check
    if (!vol)
    {
        errno = ENODEV;
        return -1;
    }
    if (!old_path || !new_path)
    {
        errno = EINVAL;
        return -1;
    }

    // Get the unicode name for the entry and find its parent directory
    // TODO: This looks horrible, clean it up
    dir = strdup(new_path);
    if (!dir)
    {
        errno = EINVAL;
        goto cleanup;
    }
    name = strrchr(dir, '/');
    if (name)
        name++;
    else
        name = dir;
    uname_len = ntfs_mbstoucs(name, &uname);
    if (uname_len < 0) {
        errno = EINVAL;
        goto cleanup;
    }
    *name = 0;

    // Find the entry
    ni = ntfs_inode_open_p(vol, old_path);
    if (!ni)
    {
        errno = ENOENT;
        res = -1;
        goto cleanup;
    }

    // Open the entries new parent directory
    dir_ni = ntfs_inode_open_p(vol, dir);
    if (!dir_ni)
    {
        errno = ENOENT;
        res = -1;
        goto cleanup;
    }

    // Link the entry to its new parent
    if (ntfs_link(ni, dir_ni, uname, uname_len))
    {
        res = -1;
        goto cleanup;
    }

    // Update entry times
    ntfs_inode_touch(dir_ni, NTFS_UPDATE_MCTIME);

    // Sync the entry to disc
    ntfs_inode_sync(ni);

cleanup:

    if(dir_ni)
        ntfs_inode_close(dir_ni);
    if(ni)
        ntfs_inode_close(ni);
    if(uname)
        free(uname);
    if(dir)
        free(dir);
    return res;
}

int ntfs_unlink_p(ntfs_volume *vol, const char *path)
{
    ntfs_inode *dir_ni = NULL, *ni = NULL;
    char *dir = NULL;
    char *name = NULL;
    ntfschar *uname = NULL;
    int uname_len;
    int res = 0;

    // Sanity check
    if (!vol)
    {
        errno = ENODEV;
        return -1;
    }
    if (!path)
    {
        errno = EINVAL;
        return -1;
    }

    // Get the unicode name for the entry and find its parent directory
    // TODO: This looks horrible
    dir = strdup(path);
    if (!dir) {
        errno = EINVAL;
        goto cleanup;
    }
    name = strrchr(dir, '/');
    if (name)
        name++;
    else
        name = dir;
    uname_len = ntfs_mbstoucs(name, &uname);
    if (uname_len < 0) {
        errno = EINVAL;
        goto cleanup;
    }
    name = strrchr(dir, '/');
    if(name)
    {
        name++;
        name[0] = 0;
    }

    // Find the entry
    ni = ntfs_inode_open_p(vol, path);
    if (!ni)
    {
        errno = ENOENT;
        res = -1;
        goto cleanup;
    }

    // Open the entries parent directory
    dir_ni = ntfs_inode_open_p(vol, dir);
    if (!dir_ni)
    {
        errno = ENOENT;
        res = -1;
        goto cleanup;
    }

    // Unlink the entry from its parent
    if (ntfs_delete(vol, path, ni, dir_ni, uname, uname_len))
        res = -1;

    // ntfs_delete() ALWAYS closes ni and dir_ni; so no need for us to anymore
    dir_ni = ni = NULL;

cleanup:

    if(dir_ni)
        ntfs_inode_close(dir_ni);
    if(ni)
        ntfs_inode_close(ni);
    if(uname)
        free(uname);
    if(dir)
        free(dir);
    return res;
}

int ntfs_inode_stat(ntfs_inode *ni, struct stat *st)
{
    // Sanity check
    if (!ni) 
    {
        errno = ENOENT;
        return -1;
    }

    // Short circuit cases were we don't actually have to do anything
    if (!st)
        return 0;

    // Zero out the stat buffer
    memset(st, 0, sizeof(struct stat));

    // Is this entry a directory
    if (ni->mrec->flags & MFT_RECORD_IS_DIRECTORY)
    {
        st->st_mode = S_IFDIR | 0777;
        st->st_nlink = 1;

        // Open the directories index allocation table attribute
        ntfs_attr *na = ntfs_attr_open(ni, AT_INDEX_ALLOCATION, NTFS_INDEX_I30, 4);
        if (na)
        {
            st->st_size = na->data_size;
            st->st_blocks = na->allocated_size >> 9;
            ntfs_attr_close(na);
        }
    } else
    {
		// it must be a file
        st->st_mode = S_IFREG | 0777;
        st->st_size = ni->data_size;
        st->st_blocks = (ni->allocated_size + 511) >> 9;
        st->st_nlink = le16_to_cpu(ni->mrec->link_count);
    }

    // Fill in the generic entry stats
    st->st_ino = ni->mft_no;
    st->st_atime = ni->last_access_time;
    st->st_ctime = ni->last_mft_change_time;
    st->st_mtime = ni->last_data_change_time;

    // Update entry times
    ntfs_inode_touch(ni, NTFS_UPDATE_ATIME);

    return 0;
}

void ntfs_inode_touch(ntfs_inode *ni, ntfs_time_update_flags mask)
{
#ifdef ATIME_DISABLED	
    // Run the access time update strategy against the device driver settings first
    mask &= ~NTFS_UPDATE_ATIME;
#endif // ATIME_DISABLED

    // Update entry times
    if (mask)
        ntfs_inode_update_times(ni, mask);
}

int ntfs_ucstombs_with_il(const ntfschar *ins, const int ins_len, char **outs, int outs_len)
{
    // Sanity check
    if (!ins || !ins_len || !outs)
        return 0;

    char *ucstombs_out = NULL;

    // Convert the unicode string to our current local
    int len = ntfs_ucstombs(ins, ins_len, &ucstombs_out, outs_len);
    if (len == -1)
    {
		if (ucstombs_out)
			free(ucstombs_out);
		
		if (errno == EILSEQ)
		{
			// The string could not be converted to the current local,
			// do it manually by replacing non-ASCII characters with underscores
			if (outs_len >= ins_len)
			{
				*outs = (char*)malloc(ins_len + 1);
				if (!*outs)
					goto NO_MEM;
				
				for (int i = 0; i < ins_len; i++)
				{
					ntfschar uc = le16_to_cpu(ins[i]);
					if (uc > 0xff)
						uc = (ntfschar)'_';
					*outs[i] = (char)uc;
				}
				*outs[ins_len] = '\0';
				len = ins_len;
			}
		}
    } else if (ucstombs_out)
    {
        // decrease using memory
        *outs = (char *)realloc(ucstombs_out, strlen(ucstombs_out) + 1);
        if(!*outs)
            goto NO_MEM;
    }

    return len;

NO_MEM:
	errno = ENOMEM;
	return -1;
}
