/**
 * ntfs_dir.c - devoptab directory routines for NTFS-based devices.
 *
 * Copyright (c) 2021 Anton "Ant-ON"
 * Copyright (c) 2010 Dimok
 * Copyright (c) 2009 Rhys "Shareese" Koedijk
 * Copyright (c) 2006 Michael "Chishm" Chisholm
 *
 * This program/include file is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program/include file is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdlib.h>
#include <sys/statvfs.h>
#include <sys/stat.h>
#include <errno.h>
#include <string.h>

#include "ntfs.h"
#include "ntfsdir.h"

#include <inode.h>
#include <dir.h>

int ntfs_stat_r(ntfs_volume *vol, const char *path, struct stat *st)
{
    // Short circuit cases were we don't actually have to do anything
    if (!st || !path)
        return 0;

    ntfs_log_trace("path %s, st %p\n", path, st);

    ntfs_inode *ni = NULL;

	// Sanity check
    if (!vol) {
        errno = ENODEV;
        return -1;
    }

    if(strcmp(path, ".") == 0 || strcmp(path, "..") == 0)
    {
        memset(st, 0, sizeof(struct stat));
        st->st_mode = S_IFDIR;
        return 0;
    }

    // Find the entry
    ni = ntfs_inode_open_p(vol, path);
    if (!ni) {
        return -1;
    }

    // Get the entry stats
    int ret = ntfs_inode_stat(ni, st);

    // Close the entry
    ntfs_inode_close(ni);

    return ret;
}

int ntfs_link_r(ntfs_volume *vol, const char *existing, const char *newLink)
{
    ntfs_log_trace("existing %s, newLink %s\n", existing, newLink);

    // Sanity check
    if (!vol)
    {
        errno = ENODEV;
        return -1;
    }

    // Create a symbolic link between the two paths
    ntfs_inode *ni = ntfs_inode_create_p(vol, existing, S_IFLNK, newLink);
    if (!ni)
        return -1;

    // Close the symbolic link
    ntfs_inode_close(ni);

    return 0;
}

int ntfs_unlink_r(ntfs_volume *vol, const char *path)
{
    ntfs_log_trace("path %s\n", path);

    // Unlink the entry
    return ntfs_unlink_p(vol, path);
}

int ntfs_rename_r(ntfs_volume *vol, const char *oldName, const char *newName)
{
    ntfs_log_trace("oldName %s, newName %s\n", oldName, newName);

    // Sanity check
    if (!vol)
    {
        errno = ENODEV;
        return -1;
    }

    // Check that there is no existing entry with the new name
    ntfs_inode *ni = ntfs_inode_open_p(vol, newName);
    if (ni)
    {
        ntfs_inode_close(ni);
        errno = EEXIST;
        return -1;
    }

    // Link the old entry with the new one
    if (ntfs_link_p(vol, oldName, newName))
        return -1;

    // Unlink the old entry
    if (ntfs_unlink_p(vol, oldName))
    {
        ntfs_unlink_p(vol, newName);
        return -1;
    }

    return 0;
}

int ntfs_mkdir_r(ntfs_volume *vol, const char *path, int mode)
{
    ntfs_log_trace("path %s, mode %i\n", path, mode);

    // Sanity check
    if (!vol)
    {
        errno = ENODEV;
        return -1;
    }

    // Create the directory
    ntfs_inode *ni = ntfs_inode_create_p(vol, path, S_IFDIR, NULL);
    if (!ni)
        return -1;

    // Close the directory
    ntfs_inode_close(ni);

    return 0;
}

int ntfs_statvfs_r(ntfs_volume *vol, struct statvfs *buf)
{
    ntfs_log_trace("buf %p\n", buf);

    s64 size;
    int delta_bits;

    // Sanity check
    if (!vol)
    {
        errno = ENODEV;
        return -1;
    }

    // Short circuit cases were we don't actually have to do anything
    if (!buf)
        return 0;

    // Zero out the stat buffer
    memset(buf, 0, sizeof(struct statvfs));

    if(ntfs_volume_get_free_space(vol) < 0)
        return -1;

    // File system block size
    buf->f_bsize = vol->cluster_size;

    // Fundamental file system block size
    buf->f_frsize = vol->cluster_size;

    // Total number of blocks on file system in units of f_frsize
    buf->f_blocks = vol->nr_clusters;

    // Free blocks available for all and for non-privileged processes
    size = max(vol->free_clusters, 0);
    buf->f_bfree = buf->f_bavail = size;

    // Free inodes on the free space
    delta_bits = vol->cluster_size_bits - vol->mft_record_size_bits;
    if (delta_bits >= 0)
        size <<= delta_bits;
    else
        size >>= -delta_bits;

    // Number of inodes at this point in time
    buf->f_files = (vol->mftbmp_na->allocated_size << 3) + size;

    // Free inodes available for all and for non-privileged processes
    size += vol->free_mft_records;
    buf->f_ffree = buf->f_favail = max(size, 0);

    // Bit mask of f_flag values.
    buf->f_flag = (NVolReadOnly(vol) ? ST_RDONLY : 0);

    // Maximum length of filenames
    buf->f_namemax = NTFS_MAX_NAME_LEN;

    return 0;
}

/**
 * PRIVATE: Callback for directory walking
 */
int ntfs_readdir_filler(ntfs_dir_state *dir, const ntfschar *name, const int name_len, const int name_type,
                         const s64 pos, const MFT_REF mref, const unsigned dt_type)
{
    ntfs_dir_entry *entry = NULL;
    char *entry_name = NULL;

    // Sanity check
    if (!dir || !dir->vol)
    {
        errno = EINVAL;
        return -1;
    }

    // Ignore DOS file names
    if (name_type == FILE_NAME_DOS)
        return 0;

    // Preliminary check that this entry can be enumerated (as described by the volume descriptor)
    if (MREF(mref) == FILE_root || MREF(mref) >= FILE_first_user)
    {
        // Convert the entry name to our current local
        if (ntfs_ucstombs_with_il(name, name_len, &entry_name, 0) < 0)
            return -1;

        // If this is not the parent or self directory reference
        if ((strcmp(entry_name, ".") == 0) || (strcmp(entry_name, "..") == 0))
        {
        	free(entry_name);
        	return 0;
        }

        // Allocate a new directory entry
        entry = (ntfs_dir_entry *) malloc(sizeof(ntfs_dir_entry));
        if (!entry)
        {
        	free(entry_name);
            return -1;
        }

        // Setup the entry
        entry->name = entry_name;
        entry->next = NULL;
        entry->mref = MREF(mref);

        // Link the entry to the directory
        if (!dir->first)
        {
            dir->first = entry;
        } else
        {
            ntfs_dir_entry *last = dir->first;
            while (last->next) last = last->next;
            last->next = entry;
        }

    }

    return 0;
}

ntfs_dir_state *ntfs_diropen_r(ntfs_volume *vol, ntfs_dir_state *dir, const char *path)
{
    ntfs_log_trace("dir %p, path %s\n", dir, path);

    s64 position = 0;

    // Sanity check
    if (!vol)
    {
        errno = ENODEV;
        return NULL;
    }

    // Find the directory
    dir->ni = ntfs_inode_open_p(vol, path);
    if (!dir->ni)
    {
        errno = ENOENT;
        return NULL;
    }

    // Ensure that this directory is indeed a directory
    if (!(dir->ni->mrec->flags & MFT_RECORD_IS_DIRECTORY))
    {
        ntfs_inode_close(dir->ni);
        errno = ENOTDIR;
        return NULL;
    }

    // Read the directory
    dir->vol = vol;
    dir->first = dir->current = NULL;
    if (ntfs_readdir(dir->ni, &position, dir, (ntfs_filldir_t)ntfs_readdir_filler))
    {
        ntfs_dirclose_r(dir);
        return NULL;
    }

    // Move to the first entry in the directory
    dir->current = dir->first;

    // Update directory times
    ntfs_inode_touch(dir->ni, NTFS_UPDATE_ATIME);

    return dir;
}

int ntfs_dirreset_r(ntfs_dir_state *dir)
{
    ntfs_log_trace("dir %p\n", dir);

    // Sanity check
    if (!dir || !dir->vol || !dir->ni)
    {
        errno = EBADF;
        return -1;
    }

    // Move to the first entry in the directory
    dir->current = dir->first;

    // Update directory times
    ntfs_inode_touch(dir->ni, NTFS_UPDATE_ATIME);

    return 0;
}

int ntfs_dirnext_r(ntfs_dir_state *dir, char *filename, struct stat *filestat)
{
    ntfs_log_trace("dir %p, filename %p, filestat %p\n", dir, filename, filestat);

    // Sanity check
    if (!dir || !dir->vol || !dir->ni)
    {
        errno = EBADF;
        return -1;
    }

    // Check that there is a entry waiting to be fetched
    if (!dir->current)
    {
        errno = ENOENT;
        return -1;
    }

    // Fetch the current entry
    strcpy(filename, dir->current->name);
    if(filestat != NULL)
    {
		ntfs_inode *ni = ntfs_pathname_to_inode(dir->vol, dir->ni, dir->current->name);
		if (ni)
		{
			ntfs_inode_stat(ni, filestat);
			ntfs_inode_close(ni);
		}
    }

    // Move to the next entry in the directory
    dir->current = dir->current->next;

    // Update directory times
    ntfs_inode_touch(dir->ni, NTFS_UPDATE_ATIME);

    return 0;
}

int ntfs_dirclose_r(ntfs_dir_state *dir)
{
    ntfs_log_trace("dir %p\n", dir);

    // Sanity check
    if (!dir || !dir->vol)
    {
        errno = EBADF;
        return -1;
    }

    // Close the directory
    // Free the directory entries (if any)
    while (dir->first)
    {
        ntfs_dir_entry *next = dir->first->next;
        free(dir->first->name);
        free(dir->first);
        dir->first = next;
    }

    // Close the directory (if open)
    if (dir->ni)
        ntfs_inode_close(dir->ni);

    // Reset the directory state
    dir->ni = NULL;
    dir->first = NULL;
    dir->current = NULL;

    return 0;
}
